@@ -22,15 +22,15 @@
 #define THIS (*(data_t **) &(instance->pdata))
 #define FEND(X,Y) (strchr(X,Y)?strchr(X,Y):(X)+strlen(X))
 
-static char *helper_fname;
-static struct handle *first_handle;
-static struct handle **prev_handlep;
-static struct handle **handlep;
-static char tempfile;
-
 /*****************************************************************************
  * Type declarations
  *****************************************************************************/
+typedef struct argument
+{
+  char *var;
+  char *value;
+} argument_t;
+
 typedef struct data
 {
      Display *display;
@@ -44,6 +44,9 @@ typedef struct data
      uint16 mode;
      char *mimetype;
      char *href;
+     char autostart;
+     int num_arguments;
+     struct argument *args;
 } data_t;
 
 typedef struct mimetype
@@ -73,6 +76,16 @@ typedef struct
      int flag;
 } options_t;
 
+/*****************************************************************************
+ * Global variables
+ *****************************************************************************/
+static char *helper_fname;
+static char *controller_fname;
+static struct handle *first_handle;
+static struct handle **prev_handlep;
+static struct handle **handlep;
+static char tempfile;
+
 /*****************************************************************************
  * Wrapper for fork() which handles some magic needed to prevent netscape
  * from freaking out. It also prevents interferrance from signals.
@@ -84,7 +97,7 @@ static int my_fork(NPP instance)
 
      /* Mask all the signals to avoid being interrupted by a signal */
      sigfillset(&set);
-     sigprocmask(SIG_SETMASK,&set,&oset);
+     sigprocmask(SIG_SETMASK, &set, &oset);
 
      D(">>>>>>>>Forking<<<<<<<<,\n");
 
@@ -112,31 +125,48 @@ static int my_fork(NPP instance)
      }
 
      /* Restore the signal mask */
-     sigprocmask(SIG_SETMASK,&oset,&set);
+     sigprocmask(SIG_SETMASK, &oset, &set);
 
      return pid;
 }
 
+/*****************************************************************************
+ * Wrapper for putenv().
+ *****************************************************************************/
+#define ENV_BUFFER 16348
+static void my_putenv(char *buffer, int *offset, const char *var, const char *value)
+{
+     int l = strlen(var) + strlen(value) + 2;
+     if (!(*offset + l < ENV_BUFFER))
+     {
+	  D("Buffer overflow in putenv(%s=%s)\n", var, value);
+	  return;
+     }
+     
+     sprintf(buffer+*offset, "%s=%s", var, value);
+     putenv(buffer+*offset);
+     *offset += l;
+}
+
 /*****************************************************************************
  * Wrapper for execlp() that calls the helper.
  *****************************************************************************/
 static void run(NPP instance, const char *file)
 {
-     char buffer[1024];
+     char buffer[ENV_BUFFER];
+     char foo[64];
+     int offset = 0;
+     int i;
+
+     if ((THIS->flags & H_CONTROLS) && (THIS->mode == NP_FULL))
+	  THIS->flags &= ~(H_CONTROLS | H_SWALLOW | H_FILL);
 
-     if (((THIS->flags & H_SWALLOW) || (THIS->flags & H_EVENT_SWALLOW))
-	 && (THIS->flags & H_NOFULL) && (THIS->mode == NP_FULL))
-     {
-	  THIS->flags &=~ H_SWALLOW;
-	  THIS->flags &=~ H_EVENT_SWALLOW;
-     }
-     
      snprintf(buffer, sizeof(buffer), "%d,%d,%lu,%d,%d,%d,%d",
 	     THIS->flags,
 	     THIS->repeats,
 	     (unsigned long int) THIS->windata.window,
 	     (int) THIS->windata.x,
-	     (int) THIS->windata.y,
+	      (int) THIS->windata.y,
 	     (int) THIS->windata.width,
 	     (int) THIS->windata.height);
 
@@ -147,17 +177,44 @@ static void run(NPP instance, const char *file)
        THIS->displayname,
        THIS->command,
        THIS->mimetype);
+
+     offset = strlen(buffer)+1;
+
+     snprintf(foo, sizeof(foo), "%lu", (long unsigned)THIS->windata.window);
+     my_putenv(buffer, &offset, "window", foo);
+
+     snprintf(foo, sizeof(foo), "0x%lx", (long unsigned)THIS->windata.window);
+     my_putenv(buffer, &offset, "hexwindow", foo);
+
+     snprintf(foo, sizeof(foo), "%ld", (long)THIS->repeats);
+     my_putenv(buffer, &offset, "repeat", foo);
+
+     snprintf(foo, sizeof(foo), "%ld", (long)THIS->windata.width);
+     my_putenv(buffer, &offset, "width", foo);
+
+     snprintf(foo, sizeof(foo), "%ld", (long)THIS->windata.height);
+     my_putenv(buffer, &offset, "height", foo);
+
+     my_putenv(buffer, &offset, "mimetype", THIS->mimetype);
      
-     execlp(helper_fname,
-	    helper_fname,
-	    buffer,
-	    file ? file : "",
-	    THIS->winname ? THIS->winname : "",
-	    THIS->displayname ? THIS->displayname : "",
-	    THIS->command,
-	    THIS->mimetype,
-	    0);
+     my_putenv(buffer, &offset, "file", file);
+
+     my_putenv(buffer, &offset, "autostart", THIS->autostart ? "1" : "0");
+      
+     if (THIS->winname)
+	  my_putenv(buffer, &offset, "winname", THIS->winname);
      
+     if (THIS->displayname)
+	  my_putenv(buffer, &offset, "DISPLAY", THIS->displayname);
+
+     if (controller_fname)
+	  my_putenv(buffer, &offset, "controller", controller_fname);
+
+     for (i = 0; i < THIS->num_arguments; i++)
+	  my_putenv(buffer, &offset, THIS->args[i].var, THIS->args[i].value);
+
+     execlp(helper_fname, helper_fname, buffer, THIS->command, NULL);
+
      D("EXECLP FAILED!\n");
      
      _exit(EX_UNAVAILABLE);
@@ -284,14 +341,14 @@ static int parse_options(char **x, command_t *commandp)
 	  { "repeat", 		H_REPEATCOUNT 	},
 	  { "loop", 		H_LOOP 		},
 	  { "stream", 		H_STREAM 	},
-	  { "many", 		H_MANY 		},
 	  { "ignore_errors",	H_IGNORE_ERRORS },
-	  { "exits", 		H_EXITS 	},
+	  { "exits", 		H_DAEMON 	},
 	  { "nokill", 		H_DAEMON 	},
 	  { "maxaspect",	H_MAXASPECT 	},
 	  { "fill",		H_FILL 		},
-	  { "nofull",		H_NOFULL 	},
 	  { "noisy",		H_NOISY 	},
+	  { "embed",            H_EMBED         },
+	  { "noembed",          H_NOEMBED       },
 	  { NULL, 		0 		}
      };
      options_t *opt;
@@ -314,6 +371,13 @@ static int parse_options(char **x, command_t *commandp)
 	       if ((*x = gobble_fct(*x, "event_swallow", commandp, H_EVENT_SWALLOW)))
 		    return 1;
 	  }
+	  if (gobble(*x, "controls"))
+	  {
+	       commandp->flags |= H_CONTROLS | H_SWALLOW | H_FILL;
+	       commandp->winname = "mozplugger-controller";
+	       *x += 8;
+	       return 1;
+	  }
      }
      return 0;
 }
@@ -494,7 +558,7 @@ static int find_helper_file(char *basename, int (*cb)(char *,void *data), void *
      if (cb(fname, data)) return 1;
      snprintf(fname, sizeof(fname), "/usr/etc/%s", basename);
      if (cb(fname, data)) return 1;
-     snprintf(fname, sizeof(fname), "/usr/local/etc/%s", basename);
+     snprintf(fname, sizeof(fname), "/usr/local/mozilla/%s", basename);
      if (cb(fname, data)) return 1;
      snprintf(fname, sizeof(fname), "/usr/local/netscape/%s", basename);
      if (cb(fname, data)) return 1;
@@ -520,6 +584,14 @@ static int find_plugger_helper_cb(char *fname, void *data)
      return 1;
 }
 
+static int find_plugger_controller_cb(char *fname, void *data)
+{
+  struct stat buf;
+  if(stat(fname, &buf)) return 0;
+  controller_fname = strdup(fname);
+  return 1;
+}
+
 /*****************************************************************************
  * Find configuration file and read it into memory.
  *****************************************************************************/
@@ -535,8 +607,11 @@ static void do_read_config(void)
 	  return;
      }
 
-     if (find("mozplugger")) helper_fname = "mozplugger";
-     find_helper_file("mozplugger", find_plugger_helper_cb, 0);
+     if (find("mozplugger-helper")) helper_fname = "mozplugger-helper";
+     find_helper_file("mozplugger-helper", find_plugger_helper_cb, 0);
+     
+     if (find("mozplugger-controller")) controller_fname = "mozplugger-controller";
+     find_helper_file("mozplugger-controller", find_plugger_controller_cb, 0);
 
      if (!helper_fname)
      {
@@ -578,6 +653,7 @@ static int find_command(NPP instance, int streaming)
      handle_t *h;
      mimetype_t *m;
      command_t *c;
+     char embedded = (THIS->mode == NP_EMBED);
 
      D("find_command\n");
 
@@ -616,7 +692,11 @@ static int find_command(NPP instance, int streaming)
 	  {
 	       for (c = h->commands; c; c = c->next)
 	       {
-		    D("Checking command: %s\n",c->cmd);
+		    D("Checking command: %s\n", c->cmd);
+		    if (embedded && (c->flags & H_NOEMBED))
+			 continue;
+		    if ((!embedded) && (c->flags & H_EMBED))
+			 continue;
 		    if ((c->flags & H_LOOP) && (THIS->repeats != MAXINT))
 			 continue;
 		    if ((!!streaming) != (!!(c->flags & H_STREAM)))
@@ -718,6 +798,25 @@ NPError NPP_GetValue(void *instance, NPPVariable variable, void *value)
      return err;
 }
 
+/*****************************************************************************
+ * Convert a string to an integer.
+ * The string can be true, false, yes or no.
+ *****************************************************************************/
+int my_atoi(char *s, int my_true, int my_false)
+{
+     switch (s[0])
+     {
+     case 't': case 'T': case 'y': case 'Y':
+	  return my_true;
+     case 'f': case 'F': case 'n': case 'N':
+	  return my_false;
+     case '0': case '1': case '2': case '3': case '4':
+     case '5': case '6': case '7': case '8': case '9':
+	  return atoi(s);
+     }
+     return -1;
+}
+
 /*****************************************************************************
  * Initiate another instance of mozplugger, it is important to know
  * that there might be several instances going at one time.
@@ -756,34 +855,38 @@ NPError NPP_New(NPMIMEType pluginType,
      THIS->displayname = NULL;
      THIS->pid = -1;
      THIS->repeats = 1;
+     THIS->autostart = 1;
      THIS->mode = mode;
 
      if (!(THIS->mimetype = strdup(pluginType)))
 	  return NPERR_OUT_OF_MEMORY_ERROR;
 
+     THIS->num_arguments = argc;
+     if (!(THIS->args = (argument_t *)NPN_MemAlloc(sizeof(argument_t) * argc)))
+	  return NPERR_OUT_OF_MEMORY_ERROR;
+     
      for (e = 0; e < argc; e++)
      {
 	  if (!strcasecmp("loop", argn[e]))
 	  {
-	       switch(argv[e][0])
-	       {
-	       case 't': case 'T': case 'y': case 'Y':
-		    THIS->repeats = MAXINT;
-		    break;
-	       case 'f': case 'F': case 'n': case 'N':
-		    THIS->repeats = 1;
-		    break;
-	       case '0': case '1': case '2': case '3': case '4':
-	       case '5': case '6': case '7': case '8': case '9':
-		    THIS->repeats = atoi(argv[e]);
-		    if (THIS->repeats < 1) THIS->repeats = 1;
-	       }
+	       THIS->repeats = my_atoi(argv[e], MAXINT, 1);
+	  }
+	  else if (!strcasecmp("autostart", argn[e]))
+	  {
+	       THIS->autostart = !!my_atoi(argv[e], 1, 0);
 	  }
 	  /* get the index of the src attribute */
 	  else if (!strcasecmp("src", argn[e]))
 	  {
 	       src_attr = e;
-	  }      
+	  }
+	  /* copy the option to put it into the environement later */
+	  if ((THIS->args[e].var = (char *)malloc(strlen(argn[e]) + 5)))
+	  {
+	       memcpy(THIS->args[e].var, "VAR_", 4);
+	       memcpy(THIS->args[e].var+4, argn[e], strlen(argn[e])+1);
+	       THIS->args[e].value = strdup(argv[e]);
+	  }
      }
 
      /* Special case for quicktime. If there's an href attribute, we 
@@ -807,7 +910,7 @@ NPError NPP_New(NPMIMEType pluginType,
  *****************************************************************************/
 NPError NPP_Destroy(NPP instance, NPSavedData** save)
 {
-     int status;
+     int e;
 
      D("Destroy\n");
 
@@ -820,6 +923,13 @@ NPError NPP_Destroy(NPP instance, NPSavedData** save)
  	  if (THIS->pid > 0)
 	       my_kill(-THIS->pid);
 
+	  for (e = 0; e < THIS->num_arguments; e++)
+	  {
+	       free((char *)THIS->args[e].var);
+	       free((char *)THIS->args[e].value);
+	  }
+	  NPN_MemFree((char *)THIS->args);
+
 	  free(THIS->mimetype);
 	  free(THIS->href);
 
@@ -827,10 +937,6 @@ NPError NPP_Destroy(NPP instance, NPSavedData** save)
 	  THIS = NULL;
      }
 
-     D("Waiting sons\n");
-
-     while (waitpid(-1, &status, WNOHANG) > 0);
-
      D("Destroy finished\n");
   
      return NPERR_NO_ERROR;
@@ -872,7 +978,7 @@ NPError NPP_NewStream(NPP instance,
 
      D("Mime type %s\n", type);
      D("Url is %s (seekable=%d)\n", stream->url, seekable);
-  
+     
      if (!find_command(instance, 1) && !find_command(instance, 0))
      {
 	  NPN_Status(instance, "No appropriate application found !");
@@ -880,9 +986,9 @@ NPError NPP_NewStream(NPP instance,
      }
 
      if ((THIS->flags & H_STREAM)
-	 && strncasecmp(stream->url,"file:",5)
-	 && strncasecmp(stream->url,"imap:",5)
-	 && strncasecmp(stream->url,"mailbox:",8))
+	 && strncasecmp(stream->url, "file:", 5)
+	 && strncasecmp(stream->url, "imap:", 5)
+	 && strncasecmp(stream->url, "mailbox:", 8))
      {
 	  *stype = NP_NORMAL;
 	  tempfile = 0;
@@ -948,38 +1054,46 @@ void NPP_StreamAsFile(NPP instance,
 }
 
 /*****************************************************************************
- * Wrapper for maximize_window().
+ * Find and resize the embedded window.
  *****************************************************************************/
-int maximize_child_window(int flags, Display *display, Window *window, int w, int h)
+int resize_window(int flags, Display *display, Window *window, int w, int h)
 {
      Window *useme = NULL;
      Window root, parent;
      Window *children = NULL;
      unsigned int num_children;
 
-     /* tell the child of the stored tree to resize, sadly we don't know
-	the handle of that window, since the helper handles it for us */
-     if (!XQueryTree(display, *window, &root, &parent, &children, &num_children))
+     if (((flags & H_SWALLOW) || (flags & H_EVENT_SWALLOW)) &&
+	 ((flags & H_FILL) || (flags & H_MAXASPECT)))
      {
-	  D("Querytree failed!!! using current window\n");
-	  useme = window;
-     }
-     if (children == NULL) {
-	  D("No children!!! using current window\n");
-	  useme = window;
-     } else if (num_children > 1) {
-	  D("Multiple Children, Not Resizing");
+	  /* tell the child of the stored tree to resize, sadly we don't know
+	     the handle of that window, since the helper handles it for us */
+	  if (!XQueryTree(display, *window, &root, &parent, &children, &num_children))
+	  {
+	       D("Querytree failed!!! using current window\n");
+	       useme = window;
+	  }
+	  if (children == NULL) {
+	       D("No children!!! using current window\n");
+	       useme = window;
+	  } else if (num_children > 1) {
+	       D("Multiple Children, Not Resizing");
+	       XFree(children);
+	       return 1;
+	  } else {
+	       /* use the first child */
+	       useme = &children[0];
+	  }
+
+	  if (flags & H_FILL) {
+	       fill_window(display, *useme, w, h);
+	  } else if (flags & H_MAXASPECT) {
+	       max_aspect_window(display, *useme, w, h);
+	  }
+
 	  XFree(children);
-	  return 1;
-     } else {
-	  /* use the first child */
-	  useme = &children[0];
      }
 
-     maximize_window(flags, display, *useme, w, h);
-
-     XFree(children);
-
      return 0;
 }
 
@@ -1006,7 +1120,7 @@ NPError NPP_SetWindow(NPP instance, NPWindow* window)
      THIS->displayname = XDisplayName(DisplayString(THIS->display));
      THIS->windata = *window;
      
-     if (maximize_child_window(THIS->flags, THIS->display, (Window *)window, window->width, window->height))
+     if (resize_window(THIS->flags, THIS->display, (Window *)window, window->width, window->height))
 	  return NPERR_NO_ERROR;
 
      if (THIS->pid != -1)
