@@ -36,7 +36,7 @@
 typedef unsigned short uint16;
 #endif
 #ifndef _UINT32
-#if defined(__alpha)
+#if defined(__alpha) || defined(__amd64__)
 typedef unsigned int uint32;
 #else /* __alpha */
 typedef unsigned long uint32;
@@ -46,7 +46,7 @@ typedef unsigned long uint32;
 typedef short int16;
 #endif
 #ifndef _INT32
-#if defined(__alpha)
+#if defined(__alpha) || defined(__amd64__)
 typedef int int32;
 #else /* __alpha */
 typedef long int32;
